const GLib = imports.gi.GLib;
const Gio = imports.gi.Gio;
const St = imports.gi.St;
const GMenu = imports.gi.GMenu;
const GObject = imports.gi.GObject;
const Clutter = imports.gi.Clutter;
const Shell = imports.gi.Shell;
const Gtk = imports.gi.Gtk;
const Pango = imports.gi.Pango;

const Main = imports.ui.main;
const OverviewControls = imports.ui.overviewControls;
const WorkspacesView = imports.ui.workspacesView;
const Favorites = imports.ui.appFavorites;
const AppDisplay = imports.ui.appDisplay;
const ExtensionUtils = imports.misc.extensionUtils;

const Lang = imports.lang;
const Gettext = imports.gettext;
const Mainloop = imports.mainloop;

const Tweener = imports.tweener.tweener;

const SCHEMA = 'org.gnome.shell.extensions.activityAppLauncher';

var _ = Gettext.domain("activityAppLauncher").gettext;
Gettext.bindtextdomain("activityAppLauncher", ExtensionUtils.getCurrentExtension().path + "/locale");

// Version 11

var init = function() {
    return new ActivityAppLauncher();
}

class ActivityAppLauncher {

    constructor() {
        this.isEnabled = false;
        this._startupPreparedId = 0;
        this._applicationsButtons = null;
        this.appsInnerContainer = null;
        this.selected = null;
        this.currentActor = null;
        this.visibility = "atexttoensurethateverythingworks";

        this._LABELSHOWTIME	= 15/100;
        this._LABELHIDETIME = 10/100;
        this._SLIDETIME = 15/100;
        this._HOVERDELAY = 300;
        this._TITLE = true;
        this._APPDESCRIPTION = true;
        this._BORDERS = true;
    }

    _onChangedSetting() {
        this.icon_size = this._settings.get_int("icon-size");
        this.show_favorites = this._settings.get_boolean("show-favorites");
        this.show_frequent = this._settings.get_boolean("show-frequent");
        this.icon_width = Math.floor(this.icon_size * 2);
        this.icon_height = Math.floor(this.icon_size * 1.8);
        this._TITLE = this._settings.get_boolean("show-tooltips");
        this._APPDESCRIPTION = this._settings.get_boolean("show-tooltips");
        this._BORDERS = this._settings.get_boolean("show-tooltip-border");
    }


    enable() {
        // Wait until startup completed
        if (Main.layoutManager._startingUp) {
            this._startupPreparedId = Main.layoutManager.connect('startup-complete', () => {
                this._doEnable(true);
            });
        } else {
            this._doEnable(false);
        }
    }

    _doEnable(removeStartup) {
        if (removeStartup) {
            Main.layoutManager.disconnect(this._startupPreparedId);
            this._startupPreparedId = 0;
        }
        this._appSys = Shell.AppSystem.get_default();
        this._settings = ExtensionUtils.getSettings(SCHEMA);
        this._controls = Main.overview._overview._controls;
        this._settingsId = this._settings.connect('changed', () => {this._onChangedSetting();});
        this._pageChangedId = this._controls.viewSelector.connect('page-changed', () => {this._onPageChanged();});
        this._onChangedSetting();
        this.isEnabled = true;
        this.fill_elements(true);
        this._favorites = Favorites.getAppFavorites();
        this._usage = Shell.AppUsage.get_default();
        this.showingId = Main.overview.connect('showing', () => {this._show();});
        this.hidingId = Main.overview.connect('hiding', () => {this._hide();});
    }

    disable() {
        if (this.isEnabled) {
            this.fill_elements(false);
            Main.overview.disconnect(this.showingId);
            Main.overview.disconnect(this.hidingId);
            this._settings.disconnect(this._settingsId);
            this._controls.viewSelector.disconnect(this._pageChangedId);
        }
        this.isEnabled = false;
    }

    _onPageChanged() {
        if ((this._controls.viewSelector._activePage == this.appsLaunchPage) ||
            (this._controls.viewSelector._activePage == this._controls.viewSelector._workspacesPage)) {
                this.appsContainer.show();
            } else {
                this.appsContainer.hide();
            }
    }

    _destroyCurrentActor() {
        if (this._allocationID) {
            GLib.source_remove(this._allocationID);
            this._allocationID = 0;
        }
        if (this.currentActor !== null) {
            this.appsLaunchContainer.remove_actor(this.currentActor);
            this.currentActor.destroy();
            this.currentActor = null;
        }
    }

    _hide() {
        this._destroyCurrentActor();
        this.appsLaunchContainer.hide();
    }

    _show() {
        this.visibility = "atexttoensurethateverythingworks";
        this.selected = null;
        this.setVisibility();
        this.appsLaunchContainer.hide();
        this._fillCategories();
    }

    _allocationChangedActivities() {
        let [stageX, stageY] = this._controls._group.get_transformed_position();
        let [stageWidth, stageHeight] = this._controls._group.get_transformed_size();
        this._stageX = stageX;
        this._stageY = stageY;
        this._stageWidth = stageWidth;
        this._stageHeight = stageHeight;
        //global.log(`X: ${stageX}x${stageY}; ${stageWidth}x${stageHeight}`);
    }

    // Modifies the Activities view to add our stage with the application categories
    fill_elements(desired_mode) {

        if (desired_mode) {
            this.appsContainer = new St.Bin({x_expand: false, y_expand: true});
            this.appsLaunchContainer = new St.Bin({});
            this._controls.viewSelector.show();
            this._controls._thumbnailsSlider.show();
            this._alloc2 = this._controls._group.connect_after("notify::allocation", () => {
                this._allocationChangedActivities();
            });
        } else {
            this._controls._group.disconnect(this._alloc2);
            this._controls._group.remove_actor(this.appsContainer);
            this._controls.viewSelector.remove_actor(this.appsLaunchPage);
            this.appsLaunchPage = null;
            this.appsContainer = null;
            this.appsLaunchContainer = null;
            this.appsInnerContainer = null;
            this._destroyCurrentActor();
        }

        this._controls._group.remove_actor(this._controls.viewSelector);
        this._controls._group.remove_actor(this._controls._thumbnailsSlider);

        if (desired_mode) {
            this._controls._group.add_child(this.appsContainer);
            this.appsLaunchPage = this._controls.viewSelector._addPage(this.appsLaunchContainer,"", '');
        }
        this._controls._group.add_child(this._controls.viewSelector);
        this._controls._group.add_actor(this._controls._thumbnailsSlider);
    }

    _fillCategories() {
        this.selected = null;
        if (this.appsInnerContainer !== null) {
            this.appsContainer.remove_actor(this.appsInnerContainer);
        }
        this.appsInnerContainer = new St.BoxLayout({ vertical: true });
        this.appsInnerContainer.x_align = Clutter.ActorAlign.FILL;
        this.appsInnerContainer.y_align = Clutter.ActorAlign.START;
        this.appsInnerContainer.x_expand = false;
        this.appsInnerContainer.y_expand = false;
        this.appsContainer.add_actor(this.appsInnerContainer);

        this.appsInnerContainer.buttons = [];
        this.appsInnerContainer.appList=[];
        this.appsInnerContainer.appClass=[];

        let tree = new GMenu.Tree({ menu_basename: 'applications.menu' });
        tree.load_sync();
        let root = tree.get_root_directory();
        let categoryMenuItem = new AALCathegory_Menu_Item(this,1,_("Windows"), null);
        this.appsInnerContainer.add_child(categoryMenuItem);
        this.appsInnerContainer.buttons.push(categoryMenuItem);

        if (this.show_favorites) {
            let favoritesMenuItem = new AALCathegory_Menu_Item(this,2,_("Favorites"), null);
            this.appsInnerContainer.add_actor(favoritesMenuItem);
            this.appsInnerContainer.buttons.push(favoritesMenuItem);
        }

        if (this.show_frequent) {
            let mostUsedMenuItem = new AALCathegory_Menu_Item(this,3,_("Frequent"), null);
            this.appsInnerContainer.add_actor(mostUsedMenuItem);
            this.appsInnerContainer.buttons.push(mostUsedMenuItem);
        }

        let iter = root.iter();
        let nextType;
        while ((nextType = iter.next()) != GMenu.TreeItemType.INVALID) {
            if (nextType == GMenu.TreeItemType.DIRECTORY) {
                let dir = iter.get_directory();
                if (!dir.get_is_nodisplay()) {
                    let childrens = this._fillCategories2(dir,[]);
                    if (childrens.length != 0) {
                        childrens.sort(this._sortApps);
                        let item = { dirItem: dir, dirChilds: childrens };
                        this.appsInnerContainer.appClass.push(item);
                    }
                }
            }
        }
        this.appsInnerContainer.appList.sort(this._sortApps);
        for (var i = 0; i < this.appsInnerContainer.appClass.length; i++) {
            let categoryMenuItem = new AALCathegory_Menu_Item(this,0,this.appsInnerContainer.appClass[i].dirItem.get_name(), this.appsInnerContainer.appClass[i].dirChilds);
            this.appsInnerContainer.add_actor(categoryMenuItem);
            this.appsInnerContainer.buttons.push(categoryMenuItem);
        }
    }

    _sortApps(param1, param2) {
        if (!param1) {
            return 1;
        }
        if (!param2) {
            return -1;
        }
        if (param1.get_name().toUpperCase()<param2.get_name().toUpperCase()) {
            return -1;
        } else {
            return 1;
        }
    }

    _fillCategories2(dir, childrens) {
        let iter = dir.iter();
        let nextType;

        while ((nextType = iter.next()) != GMenu.TreeItemType.INVALID) {
            if (nextType == GMenu.TreeItemType.ENTRY) {
                let entry = iter.get_entry();
                if (!entry.get_app_info().get_nodisplay()) {
                    let app = this._appSys.lookup_app(entry.get_desktop_file_id());
                    childrens.push(app);
                    this.appsInnerContainer.appList.push(app);
                }
            } else if (nextType == GMenu.TreeItemType.DIRECTORY) {
                childrens = this._fillCategories2(iter.get_directory(), childrens);
            }
        }
        return childrens;
    }

    _allocationChanged(button) {
        let [sizex, sizey] = this.currentActor.iconsContainer.get_size();

        if (this.last_iconx == sizex) {
            return;
        }
        this.last_iconx = sizex;
        var iconx = Math.floor(sizex / this.icon_width);
        let element_width = Math.floor(sizex / iconx);
        this.currentActor.iconsContainer.destroy_all_children();
        var position = 0;
        var currentContainer = null;
        var launcherList = null;
        switch (button.launcherType) {
            case 0:
                launcherList = button.launchers;
                break;
            case 2:
                launcherList = this._favorites.getFavorites();
                break;
            case 3:
                launcherList = this._usage.get_most_used();
                break;
            default:
                launcherList = null;
                break;
        }
        if (launcherList !== null) {
            for(let i = 0;i < launcherList.length; i++) {
                let element = launcherList[i];
                if (position == 0) {
                    currentContainer = new St.BoxLayout({vertical: false, x_expand: true});
                    currentContainer.x_expand = true;
                    currentContainer.y_expand = false;
                    this.currentActor.iconsContainer.add_child(currentContainer);
                }
                let tmpContainer = new St.BoxLayout({width: element_width, height: this.icon_height});
                tmpContainer.set_x_align(Clutter.ActorAlign.CENTER);
                tmpContainer.set_y_align(Clutter.ActorAlign.START);
                tmpContainer.set_x_expand(false);
                tmpContainer.set_y_expand(false);

                let button = new AppDisplay.AppIcon(element);
                button.set_x_align(Clutter.ActorAlign.CENTER);
                button.set_y_align(Clutter.ActorAlign.START);
                button.set_x_expand(true);
                button.set_y_expand(false);
                tmpContainer.add_child(button);
                currentContainer.add_child(tmpContainer);
                // contains tooltip data
                button.ttdata = {};
                button.ttdata._labelTimeoutId = 0;
                button.connect('enter-event', () => {
                    this.onHover(button);
                });
                button.connect('leave-event', () => {
                    this.onLeave(button);
                });
                button.connect('destroy', () => {
                    this.onLeave(button);
                });

                position++;
                if (position == iconx) {
                    position = 0;
                }
            }
        }
        this.appsLaunchContainer.show();
    }

    clickedCathegory(button) {
        for(var i = 0; i < this.appsInnerContainer.buttons.length; i++) {
            var tmpbutton = this.appsInnerContainer.buttons[i];
            if (button == tmpbutton) {
                tmpbutton.checked = true;
            } else {
                tmpbutton.checked = false;
            }
        }
        if (button.launcherType == 1) {
            this.selected = null; // for Windows list, the selected button must be null
        } else {
            this.selected = button.cat;
        }
        this.setVisibility();
        this._destroyCurrentActor();
        if (this.selected !== null) {
            this.currentActor = new St.ScrollView({hscrollbar_policy: Gtk.PolicyType.NEVER});
            this.currentActor.x_expand = true;
            this.currentActor.y_expand = true;
            this.appsLaunchContainer.add_actor(this.currentActor);
            this.currentActor.iconsContainer = new St.BoxLayout({ vertical: true, x_expand: true});
            this.last_iconx = 0;
            this.currentActor.iconsContainer.connect_after("notify::allocation", () => {
                if (this._allocationID) {
                    GLib.source_remove(this._allocationID);
                }
                this._allocationID = GLib.idle_add(GLib.PRIORITY_DEFAULT, () => {
                    this._allocationChanged(button);
                    this._allocationID = 0;
                    return false;
                });
            });
            this.currentActor.iconsContainer.x_fill = true;
            this.currentActor.iconsContainer.y_fill = true;
            this.currentActor.add_actor(this.currentActor.iconsContainer);
            this.appsLaunchContainer.show();
        }
    }

    setVisibility() {
        if (this.visibility != this.selected) {
            this.visibility = this.selected;
            var viewSelector = Main.overview._overview._controls.viewSelector;
            if (this.selected === null) {
                viewSelector._showPage(viewSelector._workspacesPage);
            } else {
                viewSelector._showPage(this.appsLaunchPage);
            }
        }
    }

    onHover(actor){

        // checks if cursor is over the icon
        // it is : let's setup a toolip display
        // unless it's already set
        if (actor.ttdata._labelTimeoutId == 0) {
            // if the tooltip is already displayed (on another icon)
            // we update it, else we delay it
            if (actor.ttdata._labelShowing) {
                this._showTooltip(actor);
            } else {
                actor.ttdata._labelTimeoutId = Mainloop.timeout_add(this._HOVERDELAY, () => {
                    this._showTooltip(actor);
                    actor.ttdata._labelTimeoutId = 0;
                    return false;
                } );
            }
        }
    }

    onLeave(actor) {
        // unset label display timer if needed
        if (actor.ttdata._labelTimeoutId > 0){
            Mainloop.source_remove(actor.ttdata._labelTimeoutId);
            actor.ttdata._labelTimeoutId = 0;
        }
        if (actor.ttdata._labelShowing) {
            this._hideTooltip(actor);
            actor.ttdata._labelShowing = false;
        }
    }

    _hideTooltip(actor) {
        if (actor.ttdata._ttbox) {
            Tweener.addTween(actor.ttdata._ttbox, {
                opacity: 0,
                time: this._LABELHIDETIME,
                transition: 'easeOutQuad',
                onComplete: function() {
                    Main.uiGroup.remove_actor(actor.ttdata._ttbox);
                    actor.ttdata._ttbox = null;
                }
            });
        }
    }

    _showTooltip(actor) {

        if (!this._TITLE && !this._APPDESCRIPTION) {
            this.onLeave(actor);
            return;
        }
        let icontext = '';
        let titletext = '';
        let detailtext = '';
        let should_display = false;

        //applications overview
        icontext = actor.app.get_name();

        if (this._APPDESCRIPTION) {
            let appDescription = actor.app.get_description();
            if (appDescription) {
                detailtext = appDescription;
                should_display = true;
            }
        }

        // Decide wether to show title
        if ( this._TITLE && icontext ) {
            titletext = icontext;
            should_display = true;
        }

        // If there's something to show ..
        if ( titletext && should_display ) {

            // Create a new tooltip if needed
            if (!actor.ttdata._ttbox) {
                let css_class = this._BORDERS ? 'activityAppLauncher-tooltip-borders' : 'activityAppLauncher-tooltip';
                actor.ttdata._ttbox = new St.Bin({ style_class: css_class });
                actor.ttdata._ttlayout = new St.BoxLayout({ vertical: true });
                actor.ttdata._ttlabel = new St.Label({ style_class: 'activityAppLauncher-tooltip-title', text: titletext });
                actor.ttdata._ttdetail = new St.Label({ style_class: 'activityAppLauncher-tooltip-detail', text: detailtext });
                actor.ttdata._ttlayout.add_child(actor.ttdata._ttlabel);
                actor.ttdata._ttlayout.add_child(actor.ttdata._ttdetail);
                actor.ttdata._ttbox.add_actor(actor.ttdata._ttlayout);

                // we force text wrap on both labels
                actor.ttdata._ttlabel.clutter_text.line_wrap = true;
                actor.ttdata._ttlabel.clutter_text.line_wrap_mode = Pango.WrapMode.WORD;
                actor.ttdata._ttlabel.clutter_text.ellipsize = Pango.EllipsizeMode.NONE;
                actor.ttdata._ttdetail.clutter_text.line_wrap = true;
                actor.ttdata._ttdetail.clutter_text.line_wrap_mode = Pango.WrapMode.WORD;
                actor.ttdata._ttdetail.clutter_text.ellipsize = Pango.EllipsizeMode.NONE;

                Main.uiGroup.add_actor(actor.ttdata._ttbox);
            } else {
                actor.ttdata._ttlabel.text = titletext;
                actor.ttdata._ttdetail.text = detailtext;
            }

            if (!titletext) {
                actor.ttdata._ttlabel.hide();
            } else {
                actor.ttdata._ttlabel.show();
            }
            if (!detailtext) {
                actor.ttdata._ttdetail.hide();
            } else {
                actor.ttdata._ttdetail.show();
            }

            let [stageX, stageY] = actor.get_transformed_position();
            let [iconWidth, iconHeight] = actor.get_transformed_size();
            let actorWidth = actor.ttdata._ttbox.get_width();
            let actorHeight = actor.ttdata._ttbox.get_height();
            let y = stageY + iconHeight + 5;
            let x = stageX - Math.round((actorWidth - iconWidth)/2);
            if (x < this._stageX) {
                x = this._stageX;
            }
            if (y < this._stageY) {
                y = this._stageY;
            }
            if ((x + actorWidth) > (this._stageX + this._stageWidth)) {
                x = this._stageX + this._stageWidth - actorWidth;
            }
            if ((y + actorHeight) > (this._stageY + this._stageHeight)) {
                y = this._stageY + this._stageHeight - actorHeight;
            }

            // do not show label move if not in showing mode
            if (actor.ttdata._labelShowing) {
                Tweener.addTween(actor.ttdata._ttbox, {
                    x: x,
                    y: y,
                    time: this._SLIDETIME,
                    transition: 'easeOutQuad',
                });
            } else {
                actor.ttdata._ttbox.set_position(x, y);
                Tweener.addTween(actor.ttdata._ttbox, {
                    opacity: 255,
                    time: this._LABELSHOWTIME,
                    transition: 'easeOutQuad',
                });
                actor.ttdata._labelShowing = true;
            }
        } else {
            // No tooltip to show : act like we're leaving an icon
            this.onLeave(actor);
        }
    }
}

const AALCathegory_Menu_Item = GObject.registerClass({
    GTypeName: 'AALCathegory_Menu_Item',
}, class AALCathegory_Menu_Item extends St.Button {

    _init(topClass, type, cathegory, launchers) {
        this.topClass = topClass;
        this.cat = cathegory;
        this.launchers = launchers;
        this.launcherType = type;
        super._init({label: GLib.markup_escape_text(cathegory,-1), style_class: "world-clocks-button button activityAppLauncherButton", toggle_mode: true, can_focus: true, track_hover: true});
        this.connect("clicked", () => {
            this.topClass.clickedCathegory(this);
        });
        this.show();
    }
});
